from ._version import __version__
from . import classification
from . import tangentspace
from . import channelselection
from . import estimation
from . import spatialfilters
from . import clustering
from . import stats
from . import embedding
from . import preprocessing
from . import datasets


__all__ = [
    '__version__',
    'classification',
    'tangentspace',
    'channelselection',
    'estimation',
    'spatialfilters',
    'clustering',
    'stats',
    'embedding',
    'preprocessing',
    'datasets',
]
